// Adapted from main11.cc.
// main11.cc is a part of the PYTHIA event generator.
// It illustrates how Les Houches Event File input can be used in Pythia8.
// Copyright (C) 2015 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL version 2, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// This program uses input from Deductor, with information transfer
// in the style of a Les Houches Event File.
// It also uses FastJet.
// D.E. Soper, 16 July 2017

// To use this, first execute the system command
//
//    mkfifo jetevents.lhe
//
// We examine the inclusive jet cross section and also
// look at the transverse-min sum pT distribution.
// Cf. Eur. Phys. J. C (2014) 74:2965

#include "Pythia8/Pythia.h"
#include "fastjet/PseudoJet.hh"
#include "fastjet/ClusterSequence.hh"
#include <iostream>
#include <fstream>
#include <string>
using namespace Pythia8;
int main() {
  
  // Fastjet analysis - select algorithm and parameters
  double Rparam = 0.4;
  fastjet::Strategy               strategy = fastjet::Best;
  fastjet::RecombinationScheme    recombScheme = fastjet::E_scheme;
  fastjet::JetDefinition         *jetDef = NULL;
  jetDef = new fastjet::JetDefinition(fastjet::antikt_algorithm, Rparam,
                                      recombScheme, strategy);
  // Fastjet input
  std::vector <fastjet::PseudoJet> fjInputs;
  
  cout << "PythiaJetslhe ready to open file." << endl;

  ifstream lhe;
  lhe.open("jetevents.lhe");
  string word, line;
  int hadronA     = 123456789;
  int hadronB     = 123456789;
  double EA       = 123456789.0;
  double EB       = 123456789.0;
  int npartons    = 123456789;
  int processcode = 123456789;
  double weight   = 123456789.0;
  bool ducteventOK= false;
  int flavor      = 123456789;
  int statuscode  = 123456789;
  int mother1     = 123456789;
  int mother2     = 123456789;
  int colorL      = 123456789;
  int colorR      = 123456789;
  double Px       = 123456789.0;
  double Py       = 123456789.0;
  double Pz       = 123456789.0;
  double E        = 123456789.0;
  double mass     = 123456789.0;

  // We will get NDeductorEvents from Deductor.
  // Note that when opts.allow_color_branching = true in deductor,
  // each Deductor event can have several branches, resulting in
  // several "events" being sent to Pythia.
  int NDeductorEvents = 123456789;
  
  // Reading of initial part of jetevents.lhe
  while (lhe >> word)
  {
    if (word == "<init>") // Find the "init" section.
    {
      getline(lhe,line);  // Go to the next line.
      break;
    }
  }
  lhe >> hadronA >> hadronB >> EA >> EB; // Read beams information.
  while (lhe >> word)  // Find end of the "init" section.
  {
    if (word == "</init>")
    {
      getline(lhe,line); // Go to the next line.
      break;
    }
  }
  
  // Now we are ready to read events.

  // Generator.
  // Follow "Hadron-Level Standalone"
  Pythia pythia;
  Event& event = pythia.event;                 // Convenient shorthand.
  pythia.readString("ProcessLevel:all = off"); // The trick!
  pythia.settings.parm("Beams:idA", hadronA);
  pythia.settings.parm("Beams:idB", hadronB);
  pythia.settings.parm("Beams:eCM", EA + EB);
  pythia.readString("Next:numberCount = 100000");  // Show iEvent every numberCount events.
  pythia.readString("TimeShower:MEcorrections = off");
  
  pythia.init();                               // Initialization.
  
  // Book histograms.
  Hist transminPTdist("Transverse-min PT/Delta R",100,0.0,10.0);
  Hist jetpTdistlow("Jet cross section",38,290.0,1050.0);
  Hist jetpTdisthigh("Jet cross section",40,1050.0,5050.0);
  double jetpTbinsizelow = 20.0;
  double jetpTbinsizehigh = 100.0;
  double weightsum = 0.0;
  int iEvent = 0;
  // Read events.
  while (lhe >> word)
  {
    if (word == "</LesHouchesEvents>")
    {
      cout << "Tag </LesHouchesEvents> found at event " << iEvent << endl;
      lhe >> word;
      cout << word;
      lhe >> NDeductorEvents;
      cout << " " << NDeductorEvents << " " ;
      lhe >> word;
      cout << word << " ";
      lhe >> word;
      cout << word << endl;
      break;
    }
    
    ducteventOK = true;
    if (word == "<event>")
    {
      event.reset();     // This resets the event record & creates line 0.
      getline(lhe,line); // Go to the next line.
      iEvent++;          // Increment the event index.
      lhe >> npartons >> processcode >> weight;
      if(iEvent < 6)
      {
        cout << scientific;
        cout << "event " << iEvent << ": npartons: " << npartons << " weight: " << weight << endl;
      }
      if(!(abs(weight) < 1.0e10))
      {
        cout << "event " << iEvent << ": npartons: " << npartons << " weight: " << weight << endl;
        weight = 0.0;
        ducteventOK = false;
      }
      getline(lhe,line);  // Go to the next line.
      
      double Pxsum    = 0.0;
      double Pysum    = 0.0;
      double Pzsum    = 0.0;
      double Esum     = 0.0;
      for (int n = 1; n <= npartons; n++)
      {
        lhe >> flavor >> statuscode >> mother1 >> mother2  >> colorL >> colorR
        >> Px >> Py >> Pz >> E >> mass;
        getline(lhe,line);  // Go to the next line.
        
        Pxsum += Px;
        Pysum += Py;
        Pzsum += Pz;
        Esum  += E;
 
        if (statuscode ==  1)
        {
          statuscode =  23;
          if (flavor == 0) flavor = 21;
          
          if (iEvent < 6 || !ducteventOK)
          {
            cout << std::showpoint;
            cout << "f: " <<  setw(2) << flavor << " status: " << setw(2) << statuscode
            << " colorL: " << setw(4) << colorL << " colorR: " << setw(4) << colorR
            << " Px: " << setw(10) << Px << " Py: " << setw(10) << Py << " Pz: " << setw(11) << Pz
            << " E: " << setw(10) << E << " m2: " << setw(10) << mass*mass
            << " m2calc: " << setw(10) << E*E - Px*Px - Py*Py - Pz*Pz  << endl;
          }
          
          // Add this parton to the event record.
          event.append(flavor, statuscode, colorL, colorR, Px, Py, Pz, E, mass);
        }
      }
      if (iEvent < 6)
      {
        cout << "Pxsum = " << Pxsum << " Pysum = " << Pysum
        << " Pzsum = " << Pzsum << " Esum = " << Esum << endl << endl;
      }
      lhe >> word;
      if (word != "</event>")  cout << "lhe format error.";
    } // Done with reading event: if (word == "<event>")

    // Hadronize the event.
    if(ducteventOK)
    {
      bool pythiaeventOK = pythia.next();
      if (!pythiaeventOK)
      {
        cout << "Pythia event " << iEvent << " has a problem." << endl;
        // pythia.event.list();
      }
      if(pythiaeventOK)
      {
        // Reset Fastjet input
        fjInputs.resize(0);
        // Loop over event record to decide what to pass to FastJet
        for (int i = 0; i < pythia.event.size(); ++i)
        {
          // Final state only
          if (!pythia.event[i].isFinal())        continue;
          
          // No neutrinos
          if (pythia.event[i].idAbs() == 12 || pythia.event[i].idAbs() == 14 ||
              pythia.event[i].idAbs() == 16)     continue;
          
          // Only |eta| < 3.0
          if (fabs(pythia.event[i].eta()) > 3.0) continue;
          
          // Store as input to Fastjet
          fjInputs.push_back( fastjet::PseudoJet(pythia.event[i].px(),
                                                 pythia.event[i].py(),
                                                 pythia.event[i].pz(),
                                                 pythia.event[i].e() ) );
        }
        if (fjInputs.size() == 0)
        {
          cout << "Error: event with no final state particles" << endl;
          continue;
        }
        // Run Fastjet algorithm
        vector <fastjet::PseudoJet> inclusiveJets, sortedJets;
        fastjet::ClusterSequence clustSeq(fjInputs, *jetDef);
        
        // For the first event, print the FastJet details
        if (iEvent < 2)
        {
          cout << "Ran " << jetDef->description() << endl;
          cout << "Strategy adopted by FastJet was "
          << clustSeq.strategy_string() << endl << endl;
        }
        // Extract inclusive jets sorted by pT (note minimum pT of 90.0 GeV)
        inclusiveJets = clustSeq.inclusive_jets(90.0);
        sortedJets    = sorted_by_pt(inclusiveJets);
        // Jet PT distribution
        for (unsigned int i = 0; i < sortedJets.size(); i++)
        {
          double jety = sortedJets[i].rap();
          double jetpT = sortedJets[i].pt();
          if (abs(jety) < 2.0 && jetpT > 290.0 && jetpT < 1050.0)
          {
            jetpTdistlow.fill(jetpT, weight/jetpTbinsizelow);
          }
          if (abs(jety) < 2.0 && jetpT > 1050.0 && jetpT < 5050.0)
          {
            jetpTdisthigh.fill(jetpT, weight/jetpTbinsizehigh);
          }
        }
        
        // Transverse-min PT distribution
        if (sortedJets.size()>0)
        {
          double px = sortedJets[0].py();
          double py = - sortedJets[0].px();
          double norm = sqrt(px*px + py*py);
          double nx = px/norm;
          double ny = py/norm;
          if (norm > 300.0)
          {
            double pTsum1 = 0.0;
            double pTsum2 = 0.0;
            double pTsum = 0.0;
            for (int i = 0; i < pythia.event.size(); ++i)
            {
              if (pythia.event[i].isFinal() && pythia.event[i].isCharged())
              {
                double kx = pythia.event[i].px();
                double ky = pythia.event[i].py();
                double absk = sqrt(kx*kx + ky*ky);
                double costheta = (kx*nx + ky*ny)/absk;
                double eta = pythia.event[i].eta(); // pseudorapidity
                if ( (abs(costheta) > sqrt(3.0)/2.0) && (abs(eta) < 2.5) && (absk > 0.5))
                {
                  if( costheta > 0) pTsum1 += pythia.event[i].pT();
                  else pTsum2 += pythia.event[i].pT();
                }
              }
            } // End loop over Pythia particles.
            // Find pTsum, add to total weight and fill in histogram.
            pTsum = min(pTsum1,pTsum2); // The transverse-min pT sum
            pTsum /= 10.4719755120;     // Divide by solid angle = 10*Pi/3
            weightsum += weight;
            transminPTdist.fill(pTsum, weight);
          } // End if (norm > 300.0)
        } // End analysis of transminPTdist: if (sortedJets.size()>0)
        
      } // End if(pythiaeventOK)
    } //End if(ducteventOK)
  } // Done with reading events: while (lhe >> word).
  cout << "Done reading " << iEvent << " events sent to Pythia out of "
       << NDeductorEvents << " created by Deductor." << endl;
  lhe.close();
  
  // Give statistics. Print histograms.
  pythia.stat();
  cout << "The total weight is " << weightsum << endl;
  transminPTdist = transminPTdist/weightsum;
  cout << transminPTdist;
  transminPTdist.table();
  jetpTdistlow = jetpTdistlow/NDeductorEvents;
  cout << jetpTdistlow;
  jetpTdistlow.table();
  jetpTdisthigh = jetpTdisthigh/NDeductorEvents;
  cout << jetpTdisthigh;
  jetpTdisthigh.table();
  
  ofstream myfile;
  myfile.open ("pythiaductR04.dat");
  myfile << "# jet cross section vs jet PT (nb/GeV).\n";
  myfile << "# " << NDeductorEvents << " events. \n";
  jetpTdistlow.table(myfile);
  jetpTdisthigh.table(myfile);
  myfile.close();
  
  // Done.
  return 0;
}
