/*  File: analyser-lhe.cc
 *  Project: Deductor
 *  Created by Zoltán Nagy on 9/8/13.
 *  Modified by D. Soper on 6 July 2017
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   local headers  */
#include "analyzer-lhe.h"

void UserLHE::initfunc()
{
  /*   Creating a file to write to   */
  lhe.open("/Users/DES/pythia8226/pythiaDES/jetevents.lhe");
  lhe << "<LesHouchesEvents version=\"1.0\">" << endl;
  lhe << "<!--" << endl;
  lhe << "Les Houches event file" << endl;
  lhe << "produced by Deductor" << endl;
  lhe << "August 2017" << endl;
  lhe << "-->" << endl;
  
  myran.seed(rdseed()); // Random seed for random number generator.
  eventnumber = 0;
  MaxLHEevents = 30000000;
  // The analyzer will turn itself off after this number of calls from deductor.
  max_number_of_events = MaxLHEevents+2;
}

void UserLHE::userfunc(const shower_history<hhc>& history)
{
  std::uniform_real_distribution<double> flat_dist1(-1.0,1.0);
  
  if(eventnumber == 0)
  {
    auto& p =  history.cbegin()->state;
    rts = (p[hadron{-1}].momentum+p[hadron{0}].momentum).mag();
    hadronA = p[hadron{0}].key;
    hadronB = p[hadron{-1}].key;
    lhe << "<init>" << endl;
    lhe << showpoint;
    lhe << hadronA <<" "<< hadronB << " "
        << rts/2.0 <<" "<< rts/2.0
        << " 13100 13100 1 1 -4 1" << endl;
    lhe << "1000.0 1000.0 1000.0 1" << endl;
    lhe << "</init>" << endl;
  }
  eventnumber++; // Counts the number of Deductor event sets

  constexpr double tonb = 389379.338;
  
  if(eventnumber > 100 && eventnumber%100000 == 0)
  {
    cout << "analyzer-lhe event number " << eventnumber << endl;
  }

  /*   the shower stage after the shower evolution   */
  for(auto& showered : history)
  {
    /*   showered state   */
    auto& p = showered.state;
 
    // The weight for a probability conserving shower
    // including factor to use NLO pdfs at hard scale.
    double wStd  = showered.weights.unitary.main*color_weight(showered)*tonb;
    wStd *= showered.weights.unitary.pdf_ratio;
    // Now include the threshold factors to get the full weight.
    double wFull = wStd*showered.weights.threshold.main;
    wFull *= showered.weights.threshold.reg;
    wFull *= showered.weights.threshold.delta;
    // --------------------------------------------------
    //   Write an lhe record for each event.
    // --------------------------------------------------
    
    if (eventnumber == MaxLHEevents + 1)
    {
      lhe << "</LesHouchesEvents>" << endl;
      lhe << "<!-- "
          << MaxLHEevents <<" events written."
          << "-->" << endl;
      cout << " tag </LesHouchesEvents> written after "
           << MaxLHEevents << " Deductor events." << endl;
      lhe.close();
    }
    
    if (eventnumber <= MaxLHEevents)
    {
      unsigned int np =  p.template number_of<duct::qcd_out>();
      unsigned int npartons = np + 2; // 2 IS partons.
      
      // Assign color labels 1001, 1002, ...
      // Here flavors and colors are outgoing,
      // so opposite to flavors and colors in for incoming partons.
      
      std::vector<int> cnext(npartons,0), cprev(npartons,0);
      int c = 1001;
      int n = 1234567890;
      int flavor = 1234567890;
      double ran = flat_dist1(myran); // -1 < ran < 1
      bool colorket = true;
      if (ran < 0.0) colorket = false;
      for (unsigned int i = 0; i < npartons; i++)
      {
        flavor = p[i-1].flavor;
        if (flavor > 0) // A quark.
        {
          cprev[i] = 0;
          if (cnext[i] == 0) // Color not assigned
          {
            cnext[i] = c;
            if(colorket) n = p[i-1].color.next+1;
            else n = p[i-1].colorP.next+1;
            cprev[n] = c;
            c++;
          }
        }
        if (flavor < 0) // An antiquark.
        {
          cnext[i] = 0;
          if (cprev[i] == 0) // Color not assigned
          {
            cprev[i] = c;
            if(colorket) n = p[i-1].color.prev+1;
            else n = p[i-1].colorP.prev+1;
            cnext[n] = c;
            c++;
          }
        }
        if (flavor == 0) // A gluon.
        {
          if (cnext[i] == 0) // Color not assigned
          {
            cnext[i] = c;
            if(colorket) n = p[i-1].color.next+1;
            else n = p[i-1].colorP.next+1;
            cprev[n] = c;
            c++;
          }
          if (cprev[i] == 0) // Color not assigned
          {
            cprev[i] = c;
            if(colorket) n = p[i-1].color.prev+1;
            else n = p[i-1].colorP.prev+1;
            cnext[n] = c;
            c++;
          }
        }
      } // Done with assignment of color labels
      
      // Adjust the particle masses to match Pythia and sometimes
      // adjust the total plus and minus momentum of the final state.
      // First find the plus and minus momentum of the final state.
      double pplusFS = 0.0;  // E + Pz. No sqrt(2) in the definition.
      double pminusFS = 0.0; // E - Pz. No sqrt(2) in the definition.
      for (unsigned int i = 1; i <= np; i++)
      {
        pplusFS += p[i].momentum.plus();
        pminusFS += p[i].momentum.minus();
      }
      double pplusbeams  = rts;
      double pminusbeams = rts;
      // Adjust final state momentum if it is too large.
      // This should never happen, but it does.
      if (pplusFS > 0.98*pplusbeams) pplusFS = 0.98*pplusbeams;
      if (pminusFS > 0.98*pminusbeams) pminusFS = 0.98*pminusbeams;
      // Now we adjust the final state parton momenta so that they have
      // the right masses (and sometimes the total P+ or P- is adjusted).
      double rtshat = sqrt(pplusFS*pminusFS);
      double nplus = pplusFS/rtshat;    // A unit vector in the direction Ptot.
      double nminus = pminusFS/rtshat;  // A unit vector in the direction Ptot.
      // For these vectors, i = 1,...,np:
      std::vector<double> pdotn(np+1,0.0), pdotnmod(np+1,0.0), msq(np+1,0.0),e(np+1,0.0);
      for (unsigned int i = 1; i <= np; i++)
      {
        pdotn[i] = 0.5*(p[i].momentum.plus()*nminus + p[i].momentum.minus()*nplus);
        pdotnmod[i] = pow(pdotn[i],2);
        if (abs(p[i].flavor) == 6) pdotnmod[i] -= pow(175.0,2); // Deductor top mass.
        if (abs(p[i].flavor) == 6) msq[i] = pow(171.0,2);       // Pythia top mass.
        else if (abs(p[i].flavor) == 5) msq[i] = pow(4.8,2);    // Pythia bottom mass.
        else if (abs(p[i].flavor) == 4) msq[i] = pow(1.5,2);    // Pythia charm mass.
        // Else we keep the previous initialization msq[i] = 0.0;
      }
      // For each momentum, we scale the part orthogonal to n^\mu
      // by a factor lambda and we add an amount ei times n^\mu.
      // First, solve for scaling variable lambda.
      double lambda = 1.0;
      double dlambda = 1.0;
      int count = 0;
      while (abs(dlambda) > 0.000000001 && count < 8)
      {
        count++;
        double sumei = 0.0;
        double sumeiprime = 0.0;
        for (unsigned int i = 1; i <= np; i++)
        {
          e[i] = sqrt(pow(lambda,2)*pdotnmod[i] + msq[i]);
          sumei += e[i];
          sumeiprime += lambda*pdotnmod[i]/e[i];
        }
        dlambda = (rtshat - sumei)/sumeiprime;
        lambda += dlambda;
      }
      // We now have lambda and the e[i]. Calculate new momenta k[i];
      std::vector<duct::lorentzvector<double>> k(np+1);
      for (unsigned int i = 1; i <= np; i++)
      {
        double kiplus = lambda*(p[i].momentum.plus() - pdotn[i]*nplus) + e[i]*nplus;
        double kiminus = lambda*(p[i].momentum.minus() - pdotn[i]*nminus) + e[i]*nminus;
        k[i][1] = lambda*p[i].momentum[1];
        k[i][2] = lambda*p[i].momentum[2];
        k[i][3] = 0.5*(kiplus - kiminus);
        k[i][0] = 0.5*(kiplus + kiminus);
      }
      
      
      // ---------- Parameters for the underlying event -------------
      bool MakeUEvent = true;  // Setting this to false will kill the underlying event.
      if (pplusFS*pminusFS > 0.9*pplusbeams*pminusbeams) MakeUEvent = false;
      double pplusUE = 0.0;    // E + Pz. No sqrt(2) in the definition.
      double pminusUE = 0.0;     // E - Pz. No sqrt(2) in the definition.
      int NUEpartons = 0;
      int Ngluons = 0;
      double PTbar = 0.0;
      if (MakeUEvent)
      {
        pplusUE = 0.5*(pplusbeams - pplusFS); // Use 50% of the available p+.
        pminusUE = 0.5*(pminusbeams - pminusFS); // Use 50% of the available p-.
        double UEratio = pow((1 - pplusFS/pplusbeams)*(1 - pminusFS/pminusbeams),0.35);
        PTbar = 7.0*UEratio; // Typical PT in UE.
        constexpr int NgluonsMAX = 14;
        double pTsqtypical = 25.0;
        double yrange = log(pplusUE*pminusUE/pTsqtypical);
        Ngluons = int(yrange);  // i.e. yrange/Deltay with Deltay = 1.0;
        if (Ngluons > NgluonsMAX) Ngluons = NgluonsMAX;
        if (Ngluons < 2) Ngluons = 2;
        std::uniform_real_distribution<double> flat_dist0(0.0,1.0);
        double r = flat_dist0(myran); // 0 < r < 1
        if (r < 0.25) // This is to produce an extra soft underlying event.
        {
          Ngluons = 2;
          PTbar = 1.0;
        }
        NUEpartons = 2*Ngluons + 4;
      }
      
      // ---------- Parameters for beam remnants -------------
      // The beam remnant momenta are a little complicated in
      // the case that they have mass. (But with a 1 GeV cutoff
      // the initial state partons never have mass and so the
      // beam remnants never have mass.)
      double pplusremnants = pplusbeams - pplusFS - pplusUE;
      double pminusremnants = pminusbeams - pminusFS - pminusUE;
      double pAT = 0.5*pplusremnants;
      double pAZ = 0.5*pplusremnants;
      double pBT = 0.5*pminusremnants;
      double pBZ = -0.5*pminusremnants;
      
      if ((abs(p[-1].flavor) > 3) || (abs(p[-1].flavor) > 3))
      {
        double mAsq = 0.0;
        if (abs(p[-1].flavor) == 4) mAsq = pow(1.5,2);
        else if (abs(p[-1].flavor) == 5) mAsq = pow(4.8,2);
        double mBsq = 0.0;
        if (abs(p[0].flavor) == 4) mBsq = pow(1.5,2);
        else if (abs(p[0].flavor) == 5) mBsq = pow(4.8,2);
        
        double temp = pplusremnants*pminusremnants - mAsq - mBsq;
        temp = pow(temp,2);
        temp = temp - 4.0*mAsq*mBsq;
        temp = sqrt(temp);
        double pAplus = pplusremnants*pminusremnants + mAsq - mBsq;
        pAplus += temp;
        pAplus /= 2.0*pminusremnants;
        double pAminus = mAsq/pAplus;
        double pBminus = pplusremnants*pminusremnants - mAsq + mBsq;
        pBminus += temp;
        pBminus /= 2.0*pplusremnants;
        double pBplus = mBsq/pBminus;
        
        pAT = 0.5*(pAplus + pAminus);
        pAZ = 0.5*(pAplus - pAminus);
        pBT = 0.5*(pBplus + pBminus);
        pBZ = 0.5*(pBplus + pBminus);
      }
      
      // ---------- Write <event> line -------------
      
      lhe << showpoint;
      lhe << setprecision(20);
      lhe << "<event>" << endl;
      lhe << npartons + NUEpartons << " 1 " << wFull << " 100.0 -1.0 -1.0 " << endl;
      
      // ---------- Write beam remnant lines -------------
      
      int flavora = p[-1].flavor;
      if (flavora == 0) flavora = 21;
      lhe << setw(3) << flavora << "  1 0 0 "
      << setw(6) << cnext[0]
      << setw(6) << cprev[0]
      << setw(30) << 0.0 << setw(30) << 0.0
      << setw(30) << pAZ << setw(30) << pAT;
      if      (abs(flavora) == 4) lhe << "   1.5 1000.0 1.0" << endl;
      else if (abs(flavora) == 5) lhe << "   4.8 1000.0 1.0" << endl;
      else                        lhe << "   0.0 1000.0 1.0" << endl;
      
      int flavorb = p[0].flavor;
      if (flavorb == 0) flavorb = 21;
      lhe << setw(3) << flavorb << "  1 0 0 "
      << setw(6) << cnext[1]
      << setw(6) << cprev[1]
      << setw(30) << 0.0 << setw(30) << 0.0
      << setw(30) << pBZ << setw(30) << pBT;
      if      (abs(flavorb) == 4) lhe << "   1.5 1000.0 1.0" << endl;
      else if (abs(flavorb) == 5) lhe << "   4.8 1000.0 1.0" << endl;
      else                        lhe << "   0.0 1000.0 1.0" << endl;
      
      // -------- Write final state particle lines -----------
      
      for (unsigned int i = 1; i <= np; i++)
      {
        int flavori = p[i].flavor;
        if (flavori == 0) flavori = 21;
        lhe << setw(3) << flavori << "  1 0 0 "
        << setw(6) << cnext[i+1]
        << setw(6) << cprev[i+1]
        << setw(30) << k[i][1] << setw(30) << k[i][2]
        << setw(30) << k[i][3] << setw(30) << k[i][0];
        
        if      (abs(flavori) == 4) lhe << "   1.5 1000.0 1.0" << endl;
        else if (abs(flavori) == 5) lhe << "   4.8 1000.0 1.0" << endl;
        else if (abs(flavori) == 6) lhe << " 171.0 1000.0 1.0" << endl;
        else                        lhe << "   0.0 1000.0 1.0" << endl;
      }
      
      // -------- Generate and write out the underlying event -----------
      if (MakeUEvent) UEvent(pplusUE, pminusUE, c, Ngluons , PTbar);
      
      lhe << "</event>" << endl;
    } // End if (eventnumber < Nevents).
  } // End for(auto& showered : history)
} // End UserJets::userfunc(const shower_history<hhc>& history)
