/*  File: analyser-jets.cc
 *  Project: Deductor
 *  Created by Zoltán Nagy on 9/8/13.
 *  Modified by D. Soper on 14 October 2016
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   local headers  */
#include "analyzer-lhe.h"

#include <iomanip>    // std::setw
#include <random>     // std::normal_distribution & std::uniform_real_distribution
#include <array>      // std::array
#include <algorithm>  // std::sort

void UserLHE::UEvent(double pplusUE, double pminusUE, int c, int Ngluons, double PTbar)
{
  double PTmean = 0.0;
  std::normal_distribution<double> normal_dist(PTmean, PTbar);
  double ymax = 0.0;
  double ymin = 0.0;
  double expy = 1.0;
  double pperp = 0.0;
  constexpr int NgluonsMAX = 14;
  int Nunused = NgluonsMAX - Ngluons;
  
  // We create a color string between an up quark from hadron A
  // and an ud diquark from hadron B,
  // with some gluons in between.
  // First the upquark for hadron A.
  int qAcolor = c;
  c++;
  double pqAx = normal_dist(myran);
  double pqAy = normal_dist(myran);
  pperp = sqrt(pqAx*pqAx + pqAy*pqAy);
  ymax = log(0.3*pplusUE/pperp);   // First string gets approximately half of the pplus.
  ymin = -log(0.3*pminusUE/pperp); // First string gets approximately half of the pminus.
  expy = exp(ymax);
  double pqAplus  = pperp*expy;  // Pplus  = E + Pz with no sqrt(2)
  double pqAminus = pperp/expy;  // Pminus = E - Pz with no sqrt(2)
  // Now some gluons.
  std::array<int, NgluonsMAX> c1;
  std::array<double, NgluonsMAX> y1;
  std::array<double, NgluonsMAX> p1x;
  std::array<double, NgluonsMAX> p1y;
  std::array<double, NgluonsMAX> p1plus;
  std::array<double, NgluonsMAX> p1minus;
  std::uniform_real_distribution<double> flat_dist1(ymin,ymax);
  for (int i = 0; i < Ngluons; i++)
  {
    double ynew = flat_dist1(myran); // ymin < ynew < ymax
    y1[i] = ynew; // y[i] is new value to start.
    p1x[i] = normal_dist(myran);
    p1y[i] = normal_dist(myran);
    c1[i] = c;
    c++;
  }
  // We put the ys in order.
  std::sort(y1.begin(),y1.end()-Nunused,std::greater<double>());
  // for (int i = 0; i < Ngluons; i++) cout << "y1["<<i<<"] = "<< y1[i] << endl;
  // Now y1[0] > y1[1] > ... > y1[Ngluons-1]
  // And we have transverse momenta
  // p1x[0], p1x[1], ... , p1x[Ngluons-1]
  // p1y[0], p1y[1], ... , p1y[Ngluons-1]
  for (int i = 0; i < Ngluons; i++)
  {
    pperp = sqrt(p1x[i]*p1x[i] + p1y[i]*p1y[i]);
    expy = exp(y1[i]);
    p1plus[i] = pperp*expy;
    p1minus[i] = pperp/expy;
  }
  // Add a ud diquark for hadron B.
  double pqqBx = -pqAx;
  double pqqBy = -pqAy;
  pperp = sqrt(pqqBx*pqqBx + pqqBy*pqqBy);
  expy = exp(ymin);
  double pqqBplus = pperp*expy;
  double pqqBminus = pperp/expy;
  
  // We create a color string between a ud diquark from hadron A
  // and an  up quark from hadron B,
  // with some gluons in between.
  // Now the ud diquark for hadron A.
  int qqAcolor = c;
  c++;
  double pqqAx = normal_dist(myran);
  double pqqAy = normal_dist(myran);
  pperp = sqrt(pqqAx*pqqAx + pqqAy*pqqAy);
  ymax = log(0.3*pplusUE/pperp);
  ymin = -log(0.3*pminusUE/pperp);
  expy = exp(ymax);
  double pqqAplus  = pperp*expy;
  double pqqAminus = pperp/expy;
  // Now some gluons.
  std::array<int, NgluonsMAX> c2;
  std::array<double, NgluonsMAX> y2;
  std::array<double, NgluonsMAX> p2x;
  std::array<double, NgluonsMAX> p2y;
  std::array<double, NgluonsMAX> p2plus;
  std::array<double, NgluonsMAX> p2minus;
  std::uniform_real_distribution<double> flat_dist2(ymin,ymax);
  for (int i = 0; i < Ngluons; i++)
  {
    double ynew = flat_dist2(myran); // ymin < ynew < ymax
    y2[i] = ynew;
    p2x[i] = - p1x[i];
    p2y[i] = - p1y[i];
    c2[i] = c;
    c++;
  }
  std::sort(y2.begin(),y2.end()-Nunused,std::greater<double>());
  // for (int i = 0; i < Ngluons; i++) cout << "y2["<<i<<"] = "<< y2[i] << endl;
  // Now y2[0] > y2[1] > ... > y2[Ngluons-1]
  // And we have transverse momenta
  // p2x[0], p2x[1], ... , p2x[Ngluons-1]
  // p2y[0], p2y[1], ... , p2y[Ngluons-1]
  for (int i = 0; i < Ngluons; i++)
  {
    pperp = sqrt(p2x[i]*p2x[i] + p2y[i]*p2y[i]);
    expy = exp(y2[i]);
    p2plus[i] = pperp*expy;
    p2minus[i] = pperp/expy;
  }
  // Add a  uquark for hadron B.
  double pqBx = -pqqAx;
  double pqBy = -pqqAy;
  pperp = sqrt(pqBx*pqBx + pqBy*pqBy);
  expy = exp(ymin);
  double pqBplus  = pperp*expy;
  double pqBminus = pperp/expy;
  
  
  // Adjust the total momentum used.
  // Use a combination of an overall rescaling and a boost.
  double pplusused =   pqAplus + pqqAplus + pqBplus + pqqBplus;
  double pminusused =  pqAminus + pqqAminus + pqBminus + pqqBminus;
  for (int i = 0; i < Ngluons; i++)
  {
    pplusused  += p1plus[i] + p2plus[i];
    pminusused += p1minus[i] + p2minus[i];
  }
  double scaleplusplus = pplusUE/pplusused;
  double scaleplus  = sqrt(scaleplusplus);
  double scaleminusminus =  pminusUE/pminusused ;
  double scaleminus = sqrt(scaleminusminus);
  double scaleplusminus = scaleplus*scaleminus;
  pqAplus  *= scaleplusplus;
  pqAminus *= scaleminusminus;
  pqAx     *= scaleplusminus;
  pqAy     *= scaleplusminus;
  pqBplus  *= scaleplusplus;
  pqBminus *= scaleminusminus;
  pqBx     *= scaleplusminus;
  pqBy     *= scaleplusminus;
  pqqAplus  *= scaleplusplus;
  pqqAminus *= scaleminusminus;
  pqqAx     *= scaleplusminus;
  pqqAy     *= scaleplusminus;
  pqqBplus  *= scaleplusplus;
  pqqBminus *= scaleminusminus;
  pqqBx     *= scaleplusminus;
  pqqBy     *= scaleplusminus;
  for (int i = 0; i < Ngluons; i++)
  {
    p1plus[i]  *= scaleplusplus;
    p2plus[i]  *= scaleplusplus;
    p1minus[i] *= scaleminusminus;
    p2minus[i] *= scaleminusminus;
    p1x[i]     *= scaleplusminus;
    p2x[i]     *= scaleplusminus;
    p1y[i]     *= scaleplusminus;
    p2y[i]     *= scaleplusminus;
  }
  
  // Write out the partons
  
  // First string, from color 3 in hadron A to 3-bar in hadron B.
  int f3A = 2;                  // u for a proton
  if (hadronA < 0) f3A = -2101; // (ud)-bar for a proton-bar
  lhe << setw(5) << f3A << "  1 0 0 "
  << setw(6) << qAcolor
  << setw(6) << 0
  << setw(30) << pqAx << setw(30) << pqAy
  << setw(30) << 0.5*(pqAplus - pqAminus)
  << setw(30) << 0.5*(pqAplus + pqAminus)
  << "   0.0 1000.0 1.0" << endl;
  
  for (int i = 0; i < Ngluons; i++)
  {
    lhe << setw(5) << 21 << "  1 0 0 "
    << setw(6) << c1[i];
    if (i > 0)   lhe << setw(6) << c1[i-1];
    else         lhe << setw(6) << qAcolor;
    lhe << setw(30) << p1x[i] << setw(30) << p1y[i]
    << setw(30) << 0.5*(p1plus[i] - p1minus[i])
    << setw(30) << 0.5*(p1plus[i] + p1minus[i])
    << "   0.0 1000.0 1.0" << endl;
  }
  
  int f3barB = 2101;            // (ud) for a proton
  if (hadronB < 0) f3barB = -2; // u-bar for a proton-bar
  lhe << setw(5) << f3barB << "  1 0 0 "
  << setw(6) << 0
  << setw(6) << c1[Ngluons-1]
  << setw(30) << pqqBx << setw(30) << pqqBy
  << setw(30) << 0.5*(pqqBplus - pqqBminus)
  << setw(30) << 0.5*(pqqBplus + pqqBminus)
  << "   0.0 1000.0 1.0" << endl;
  
  // Second string, from color 3-bar in hadron A to 3 in hadron B.
  int f3barA = 2101;            // (ud) for a proton
  if (hadronA < 0) f3barA = -2; // u-bar for a proton-bar
  lhe << setw(5) << f3barA << "  1 0 0 "
  << setw(6) << 0
  << setw(6) << qqAcolor
  << setw(30) << pqqAx << setw(30) << pqqAy
  << setw(30) << 0.5*(pqqAplus - pqqAminus)
  << setw(30) << 0.5*(pqqAplus + pqqAminus)
  << "   0.0 1000.0 1.0" << endl;
  
  for (int i = 0; i < Ngluons; i++)
  {
    lhe << setw(5) << 21 << "  1 0 0 ";
    if (i > 0)   lhe << setw(6) << c2[i-1];
    else         lhe << setw(6) << qqAcolor;
    lhe << setw(6) << c2[i]
    << setw(30) << p2x[i] << setw(30) << p2y[i]
    << setw(30) << 0.5*(p2plus[i] - p2minus[i])
    << setw(30) << 0.5*(p2plus[i] + p2minus[i])
    << "   0.0 1000.0 1.0" << endl;
  }
  
  int f3B = 2;                  // u for a proton
  if (hadronB < 0) f3B = -2101; // (ud)-bar for a proton-bar
  lhe << setw(5) << f3B << "  1 0 0 "
  << setw(6) << c2[Ngluons-1]
  << setw(6) << 0
  << setw(30) << pqBx << setw(30) << pqBy
  << setw(30) << 0.5*(pqBplus - pqBminus)
  << setw(30) << 0.5*(pqBplus + pqBminus)
  << "   0.0 1000.0 1.0" << endl;
}

